@@ -5,6 +5,22 @@
 #include <R_ext/Parse.h>
 #include <R_ext/Print.h>
 
+/* R 4.0.1 broke EXTPTR_PTR ABI so re-map it to safety at
+   the small expense of speed */
+#ifdef  EXTPTR_PTR
+#undef  EXTPTR_PTR
+#endif
+#define EXTPTR_PTR(X) R_ExternalPtrAddr(X)
+/* PROT/TAG are safe so far, but just to make sure ... */
+#ifdef  EXTPTR_PROT
+#undef  EXTPTR_PROT
+#endif
+#define EXTPTR_PROT(X) R_ExternalPtrProtected(X)
+#ifdef  EXTPTR_TAG
+#undef  EXTPTR_TAG
+#endif
+#define EXTPTR_TAG(X) R_ExternalPtrTag(X)
+
 #include <stdarg.h>
 
 /* max supported # of parameters to Java methdos */
@@ -163,7 +179,7 @@ HIDE void deserializeSEXP(SEXP o) {
 	    if (go) {
 	      _dbg(rjprintf(" - succeeded: %p\n", go));
 	      /* set the deserialized object */
-	      EXTPTR_PTR(o) = (SEXP) go;
+	      R_SetExternalPtrAddr(o, go);
 	      /* Note: currently we don't remove the serialized content, because it was created explicitly using .jcache to allow repeated saving. Once this is handled by a hook, we shall remove it. However, to assure compatibility TAG is always NULL for now, so we do clear the cache if TAG is non-null for future use. */
 	      if (EXTPTR_TAG(o) != R_NilValue) {
 		/* remove the serialized raw vector */
@@ -198,6 +214,18 @@ static void strcats(sig_buffer_t *sig, const char *add) {
   sig->len += l;
 }
 
+/* call strcats() but also convert class names to JNI notation */
+static void strcats_conv(sig_buffer_t *sig, const char *add) {
+  int ol = sig->len, nl;
+  strcats(sig, add);
+  nl = sig->len;
+  while (ol < nl) {
+    if (sig->sig[ol] == '.')
+      sig->sig[ol] = '/';
+    ol++;
+  }
+}
+
 /* initialize a signature buffer */
 HIDE void init_sigbuf(sig_buffer_t *sb) {
   sb->len = 0;
@@ -345,7 +373,8 @@ static int Rpar2jvalue(JNIEnv *env, SEXP par, jvalue *jpar, sig_buffer_t *sig, i
 	addtmpo(tmpo, jpar[jvpos++].l=newBooleanArrayI(env, LOGICAL(e),LENGTH(e)));
       }
     } else if (TYPEOF(e)==VECSXP || TYPEOF(e)==S4SXP) {
-      _dbg(rjprintf(" generic vector of length %d\n", LENGTH(e)));
+      if (TYPEOF(e) == VECSXP)
+	_dbg(rjprintf(" generic vector of length %d\n", LENGTH(e)));
       if (IS_JOBJREF(e)) {
 	jobject o=(jobject)0;
 	const char *jc=0;
@@ -375,9 +404,9 @@ static int Rpar2jvalue(JNIEnv *env, SEXP par, jvalue *jpar, sig_buffer_t *sig, i
 	}
 	if (jc) {
 	  if (*jc!='[') { /* not an array, we assume it's an object of that class */
-	    strcats(sig,"L"); strcats(sig,jc); strcats(sig,";");
+	    strcats(sig,"L"); strcats_conv(sig,jc); strcats(sig,";");
 	  } else /* array signature is passed as-is */
-	    strcats(sig,jc);
+	    strcats_conv(sig,jc);
 	} else
 	  strcats(sig,"Ljava/lang/Object;");
 	jpar[jvpos++].l=o;
